from datetime import datetime
import bisect

class IntervalError(Exception):
    pass

class Interval(object):
    """Represents an interval of time"""

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __repr__(self):
        return "Interval(" + repr(self.start) + ", " + repr(self.end) + ")"

    def __str__(self):
        return "[" + str(self.start) + " -> " + str(self.end) + "]"

    def __getattr__(self, name):
        return { }

    def __setattr__(self, name, value):
        """Set attribute"""
        # TODO: If we need to manipulate file names, offsets, lengths, etc,
        # based on start and end time changing, maybe this is the right spot?
        # Or we could just disallow changing it here.
        if (type(value) is not datetime):
            raise IntervalError("Must set datetime values")
        self.__dict__[name] = value
        if (type(self.start) is type(self.end)):
            if (self.start > self.end):
                raise IntervalError("Interval start must precede interval end")

    def __cmp__(self, other):
        """Compare two intervals.  If non-equal, order determined by start then end"""
        if (self.start == other.start):
            if (self.end < other.end):
                return -1
            if (self.end > other.end):
                return 1
            return 0
        if (self.start < other.start):
            return -1
        return 1
                
    def intersects(self, other):
        """Return True if two Interval objects intersect"""
        if (not isinstance(other, Interval)):
            raise TypeError("need Interval for intersection test")
        if (self.end <= other.start or
            self.start >= other.end):
            return False
        else:
            return True

    def is_adjacent(self, other):
        """Return True if two Intervals are adjacent (same end or start)"""
        if (not isinstance(other, Interval)):
            raise TypeError("need Interval for adjacency test")
        if (self.end == other.start or
            self.start == other.end):
            return True
        else:
            return False

    def subset(self, start, end):
        """Return a new Interval that is a subset of this one"""
        # TODO: Any magic regarding file/offset/length mapping for subsets
        if (start < self.start or end > self.end):
            raise IntervalError("not a subset")
        return Interval(start, end)            

class IntervalSet(object):
    """A non-intersecting set of intervals

    Kept sorted internally"""

    def __init__(self, iterable=None):
        self.data = []
        if iterable is not None:
            if type(iterable) is Interval:
                iterable = [iterable]
            self._add_intervals(iterable)

    def __iter__(self):
        return self.data.__iter__()

    def __repr__(self):
        return "IntervalSet(" + repr(list(self.data)) + ")"

    def __cmp__(self, other):
        # compare isn't supported, they don't really have an ordering
        raise TypeError("can't compare IntervalSets with cmp()")

    def __eq__(self, other):
        """Test equality of two IntervalSets.

        Treats adjacent Intervals as equivalent to one long interval,
        so this function really tests whether the IntervalSets cover
        the same spans of time."""
        if type(other) is not IntervalSet:
            return False
        i = 0
        j = 0
        outside = True
        try:
            while True:
                if (outside):
                    # To match, we need to be finished this set
                    if (i >= len(self) and j >= len(other)):
                        return True
                    # Or the starts need to match
                    if (self[i].start != other[j].start):
                        return False
                    outside = False
                else:
                    # We can move on if the two interval ends match
                    if (self[i].end == other[j].end):
                        i += 1
                        j += 1
                        outside = True
                    else:
                        # Whichever ends first needs to be adjacent to the next
                        if (self[i].end < other[j].end):
                            if (not self[i].is_adjacent(self[i+1])):
                                return False
                            i += 1
                        else:
                            if (not other[j].is_adjacent(other[j+1])):
                                return False
                            j += 1
        except IndexError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __iadd__(self, other):
        """Inplace add -- modifies self

        This throws an exception if the regions being added intersect."""
        if type(other) is Interval:
            other = [other]
        self._add_intervals(other)
        return self
        
    def __add__(self, other):
        """Add -- returns a new object

        This throws an exception if the regions being added intersect."""
        new = IntervalSet(self)
        new += IntervalSet(other)
        return new

    def __and__(self, other):
        """Compute a new IntervalSet from the intersection of two others

        Output intervals are built as subsets of the intervals in the
        first argument (self)."""
        # If we were given a set, intersect with each interval in that set
        if type(other) is IntervalSet:
            out = IntervalSet()
            for interval in other.data:
                out += self & interval
            return out

        if type(other) is not Interval:
            raise TypeError("can't intersect with that type")

        out = IntervalSet()
        for this in self.data:
            # If there's any overlap, add the overlapping region
            if (this.end > other.start and this.start < other.end):
                out += this.subset(max(this.start, other.start),
                                   min(this.end, other.end))
        return out
    
    def _add_intervals(self, iterable):
        for element in iter(iterable):
            self._add_single_interval(element)

    def _add_single_interval(self, interval):
        if (not isinstance(interval, Interval)):
            raise TypeError("can only add Intervals")
        for existing in self.data:
            if existing.intersects(interval):
                raise IntervalError("Tried to add overlapping interval to this set")
        bisect.insort(self.data, interval)
